/*
	bfsha1 - Brute forces a SHA1 hash
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "gpucracker.h"
#include "cudacommon.h"
#include "sha1.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>

using namespace std;

GpuCracker::GpuCracker(const char *charSet, uint32_t charSetLen, uint32_t pwLen, const uint32_t *hash, const char *hashName, int gpu)
{
	if (STR_CMP_ICASE(hashName, "sha1") == 0)
	{
		if (sha1_getKernel(pwLen, &m_cracker, &m_crackerCheat))
		{
			reportError("No SHA1 GPU kernel for password length %u.", pwLen);
		}
	}
	else
	{
		reportError("No \"%s\" hashing algorithm found for GPU.", hashName);
	}

	m_charSetLen = charSetLen;
	m_pwLen = pwLen;
	m_hash[0] = hash[0];
	m_hash[1] = hash[1];
	m_hash[2] = hash[2];
	m_hash[3] = hash[3];
	m_hash[4] = hash[4];
	m_foundPw = new char[32];
	strcpy(m_foundPw, "<not found>");

	m_maxIndex = 1;
	for (uint32_t i = 2; i < pwLen; i++)
	{
		m_maxIndex *= charSetLen;
	}

	// Init GPU
	checkCudaErrors(cudaSetDevice(gpu));
	getDefaultBlocksThreads(gpu, m_blocks, m_threads);

	checkCudaErrors(cudaMallocHost((void**) &m_h_foundPw,        16));
	checkCudaErrors(cudaMalloc    ((void**) &m_d_foundPw,        16));
	checkCudaErrors(cudaMalloc    ((void**) &m_d_shared,  268 + 256));
	uint64_t *shared = new uint64_t[(268 + 256 + 7) / 8];

	// for pwLen = 4 to maxPwLen
	//   uint32_t s_shortCutShift[16]   = max(0, ceil(log2(62 ^ pwLen)) - 24);
	//   uint64_t s_shortCut[16]        = 62 ^ (pwLen - 3);
	//   float s_shortCutReciprocal[16] = 2 ^ s_shortCutShift[i] / s_shortCut[i];
	// float s_shortCutf[3]             = 62 ^ (pwLen - 3);
	uint64_t *s_shortCut           = (uint64_t*) shared;
	uint32_t *s_shortCutShift      = (uint32_t*) (s_shortCut           + 16);
	float    *s_shortCutReciprocal = (float*)    (s_shortCutShift      + 16);
	float    *s_shortCutf          = (float*)    (s_shortCutReciprocal + 16);
	uint8_t  *s_charSetsData       = (uint8_t*)  (s_shortCutf          +  3);
	uint64_t keySpace = charSetLen * charSetLen * charSetLen;
	uint64_t shortCut = 1;

	for (uint32_t i = 0; i < pwLen - 3; i++)
	{
		shortCut *= charSetLen;
		keySpace *= charSetLen;
		s_shortCutShift[i] = 64 - countLeadingZeros(keySpace);
		if (s_shortCutShift[i] < 24)
			s_shortCutShift[i]  = 0;
		else
			s_shortCutShift[i] -= 24;
		s_shortCut[i]           = shortCut;
		s_shortCutReciprocal[i] = reciprocalRoundDown(s_shortCut[i]) * ((float) (UINT64_C(1) << s_shortCutShift[i]));
	}
	s_shortCutf[0] = (float) s_shortCut[0];
	s_shortCutf[1] = (float) s_shortCut[1];
	s_shortCutf[2] = (float) s_shortCut[2];
	for (uint32_t i = 0; i < charSetLen; i++)
	{
		s_charSetsData[i] = ((uint8_t*) charSet)[i];
	}
	m_h_foundPw[0] = 0;
	m_h_foundPw[1] = 0;
	m_h_foundPw[2] = 0;
	m_h_foundPw[3] = 0;

	// copy data to device
	checkCudaErrors(cudaMemcpy(m_d_shared,  shared,      268 + 256, cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(m_d_foundPw, m_h_foundPw,        16, cudaMemcpyHostToDevice));

	// __constant__ uint32_t c_charSetLength;
	// __constant__ float    c_charSetReciprocal;
	// __constant__ float    c_charSetLen;
	// __constant__ float    c_ncharSetLen;
	cudaMemcpyToSymbol("c_charSetLength",          &charSetLen,                                  sizeof(uint32_t)); getLastCudaError("c_charSetLength");

	float c_charSetLen = reciprocalRoundDown(charSetLen);
	cudaMemcpyToSymbol("c_charSetReciprocal",      &c_charSetLen,                                sizeof(float));    getLastCudaError("c_charSetReciprocal");
	c_charSetLen = (float) charSetLen;
	cudaMemcpyToSymbol("c_charSetLen",             &c_charSetLen,                                sizeof(float));    getLastCudaError("c_charSetLen");
	c_charSetLen = -c_charSetLen;
	cudaMemcpyToSymbol("c_ncharSetLen",            &c_charSetLen,                                sizeof(float));    getLastCudaError("c_ncharSetLen");

	// __constant__ uint32_t c_a;
	// __constant__ uint32_t c_b;
	// __constant__ uint32_t c_c;
	// __constant__ uint32_t c_d;
	// __constant__ uint32_t c_e;
	// __constant__ uint32_t c_dXORe;
	// __constant__ uint32_t c_eOld;
	uint32_t ha = hash[0], hb = hash[1], hc = hash[2], hd = hash[3], he = hash[4], dXORe, heOld;

	ha -= 0x67452301;
	hb -= 0xefcdab89;
	hc -= 0x98badcfe;
	hd -= 0x10325476;
	he -= 0xc3d2e1f0;

#define ROL(x,s) (((x) << (s)) | ((x) >> (32 - (s))))
	// SHA1_II(b,c,d,e,a,    x,79);
	hc = ROL(hc, 2);
	ha -= (hc ^ (hd ^ he)) + 0xca62c1d6 + ROL(hb, 5);
	// ha -= w[79-16] ^ c_r79[i];

	// SHA1_II(c,d,e,a,b,    x,78);
	hd = ROL(hd, 2);
	hb -= 0xca62c1d6 + ROL(hc, 5);
	dXORe = hd ^ he;
	// hb -= (dXORe ^ ha) + w[78-16];

	// SHA1_II(d,e,a,b,c,    x,77);
	heOld = he = ROL(he, 2);
	hc -= 0xca62c1d6 + ROL(hd, 5);
	// hc -= (he ^ (ha ^ hb)) + (w[77-16] ^ c_r77[i]);

	// SHA1_II(e,a,b,c,d,    x,76);
	// ha = ROL(ha, 2);
	hd -= 0xca62c1d6 + ROL(he, 5);
	// hd -= (ha ^ (hb ^ hc)) + (w[76-16] ^ c_r76[i]);

	// SHA1_II(a,b,c,d,e,    x,75);
	// hb = ROL(hb, 2);
	he -= 0xca62c1d6;
	// he -= (hb ^ (hc ^ hd)) + (w[75-16] ^ c_r75[i]) + ROL(ha, 5);

	// ...

	// SHA1_II(d,e,a,b,c,    x,72);
	// he = ROL(he, 2);
	cudaMemcpyToSymbol("c_ha",    &ha,    sizeof(uint32_t)); getLastCudaError("c_ha");
	cudaMemcpyToSymbol("c_hb",    &hb,    sizeof(uint32_t)); getLastCudaError("c_hb");
	cudaMemcpyToSymbol("c_hc",    &hc,    sizeof(uint32_t)); getLastCudaError("c_hc");
	cudaMemcpyToSymbol("c_hd",    &hd,    sizeof(uint32_t)); getLastCudaError("c_hd");
	cudaMemcpyToSymbol("c_he",    &he,    sizeof(uint32_t)); getLastCudaError("c_he");
	cudaMemcpyToSymbol("c_dXORe", &dXORe, sizeof(uint32_t)); getLastCudaError("c_dXORe");
	cudaMemcpyToSymbol("c_heOld", &heOld, sizeof(uint32_t)); getLastCudaError("c_heOld");


/*
	// 41 * 4 * 192 = 61.5 KiB
	__constant__ uint32_t c_char[192];
	__constant__ uint32_t c_char_2[192];

	__constant__ uint32_t c_r17[192],   c_r20[192],   c_r23[192],   c_r26[192],   c_r29[192],   c_r31[192],   c_r32[192],   c_r33[192],   c_r35[192],   c_r37[192];
	__constant__ uint32_t c_r38[192],   c_r41[192],   c_r43[192],   c_r44[192],   c_r45[192],   c_r47[192],   c_r48[192],   c_r49[192],   c_r50[192],   c_r53[192];
	__constant__ uint32_t c_r55[192],   c_r56[192],   c_r57[192],   c_r59[192],   c_r60[192],   c_r61[192],   c_r62[192],   c_r63[192],   c_r65[192],   c_r67[192];
	__constant__ uint32_t c_r68[192],   c_r69[192],   c_r71[192],   c_r72[192],   c_r73[192],   c_r74[192],   c_r75[192],   c_r76[192],   c_r77[192],   c_r79[192];
	__constant__ uint32_t c_r17_2[192], c_r20_2[192], c_r23_2[192], c_r26_2[192], c_r29_2[192], c_r31_2[192], c_r32_2[192], c_r33_2[192], c_r35_2[192], c_r37_2[192];
	__constant__ uint32_t c_r38_2[192], c_r41_2[192], c_r43_2[192], c_r44_2[192], c_r45_2[192], c_r47_2[192], c_r48_2[192], c_r49_2[192], c_r50_2[192], c_r53_2[192];
	__constant__ uint32_t c_r55_2[192], c_r56_2[192], c_r57_2[192], c_r59_2[192], c_r60_2[192], c_r61_2[192], c_r62_2[192], c_r63_2[192], c_r65_2[192], c_r67_2[192];
	__constant__ uint32_t c_r68_2[192], c_r69_2[192], c_r71_2[192], c_r72_2[192], c_r73_2[192], c_r74_2[192], c_r75_2[192], c_r76_2[192], c_r77_2[192], c_r79_2[192];
*/
#define c_char(i)   c_wRounds[0*192*41 + 192*    0 + i]
#define c_r(j,i)    c_wRounds[0*192*41 + 192*(j+1) + i]
#define c_char_2(i) c_wRounds[1*192*41 + 192*    0 + i]
#define c_r_2(j,i)  c_wRounds[1*192*41 + 192*(j+1) + i]
	uint32_t w[64];
	uint32_t *c_wRounds = new uint32_t[2*192*41];
	uint32_t round[] = 
		{17, 20, 23, 26, 29, 31, 32, 33, 35, 37,
		 38, 41, 43, 44, 45, 47, 48, 49, 50, 53,
		 55, 56, 57, 59, 60, 61, 62, 63, 65, 67,
		 68, 69, 71, 72, 73, 74, 75, 76, 77, 79};

	for (uint32_t i = 0; i < charSetLen; i++)
	{
		c_char(i)   = ((uint32_t) charSet[i]) << 24;
		c_char_2(i) = ((uint32_t) charSet[i]) << 16;
		w[ 0] = 0;
		w[ 1] = ROL(c_char(i), 1);
		w[ 2] = 0;
		w[ 3] = 0;
		w[ 4] = ROL(w[ 1]    , 1);
		w[ 5] = 0;
		w[ 6] = 0;
		w[ 7] = ROL(w[ 4]    , 1);
		w[ 8] = 0;
		w[ 9] = w[ 4];
		w[10] = ROL(w[ 7]    , 1);
		w[11] = 0;
		w[12] = 0;
		w[13] = ROL(w[10]    , 1);
		w[14] = 0;
		w[15] = w[ 9] ^ w[10];

		for (uint32_t j = 16; j < 64; j++)
		{
			w[j] = ROL(w[j-3] ^ w[j-8] ^ w[j-14] ^ w[j-16], 1);
		}
		for (uint32_t j = 0; j < 40; j++)
		{
			c_r(j,i)   =     w[round[j]-16];
			c_r_2(j,i) = ROL(w[round[j]-16], 24);
		}
	}
	cudaMemcpyToSymbol("c_wRounds", c_wRounds, 2*192*41 * sizeof(uint32_t));
	getLastCudaError("c_wRounds");
#undef ROL

	checkCudaErrors(cudaDeviceSynchronize());

	// clean up
	delete [] c_wRounds;
	delete [] shared;
}

GpuCracker::~GpuCracker()
{
	GpuCracker::cleanUp();
}

int GpuCracker::crack(uint64_t index)
{
/*
	GpuCracker(uint8_t *charSet, uint32_t charSetLen, uint32_t *hash);
	~GpuCracker();

	void cracker(uint64_t index);

private:
	void cleanUp();

	kernel_cracker m_cracker;
	uint32_t *m_d_foundPw, *m_d_shared, m_hash[5], m_sharedSize;
	int m_blocks, m_threads;
*/
	uint64_t maxIndex      = m_maxIndex;
	uint32_t *d_shared     = m_d_shared;
	uint32_t *d_foundPw    = m_d_foundPw;
	kernel_cracker cracker = m_crackerCheat;
	uint32_t i             = 0;
	uint32_t pwLen         = m_pwLen;
	uint32_t blocks        = m_blocks;
	uint32_t threads       = m_threads;

	index /= 95 * 95;
	if (index >= maxIndex)
	{
		return 2;
	}
	cracker(blocks, threads, d_shared, d_foundPw, pwLen, 0, 95, index, maxIndex);
	getLastCudaError("Failed kernel launch.");
/*	for (; i < 95 - 12; i += 12)
	{
		cracker(blocks, threads, d_shared, d_foundPw, pwLen, i, i + 12, index, maxIndex);
		getLastCudaError("Failed kernel launch.");
	}
	if (i < 95)
	{
		cracker(blocks, threads, d_shared, d_foundPw, pwLen, i, 95, index, maxIndex);
		getLastCudaError("Failed kernel launch.");
	}*/
	checkCudaErrors(cudaMemcpy(m_h_foundPw, d_foundPw, 4, cudaMemcpyDeviceToHost));
	checkCudaErrors(cudaDeviceSynchronize());
	if (m_h_foundPw[0] != 0)
	{
		m_h_foundPw[0] = 0;
		checkCudaErrors(cudaMemcpy(d_foundPw, m_h_foundPw, 4, cudaMemcpyHostToDevice));
		m_cracker(blocks, threads, d_shared, d_foundPw, pwLen, 0, 95, index, maxIndex);
		getLastCudaError("Failed kernel launch.");
		checkCudaErrors(cudaMemcpy(m_h_foundPw, d_foundPw, 16, cudaMemcpyDeviceToHost));
		checkCudaErrors(cudaDeviceSynchronize());
		if (m_h_foundPw[0] != 0)
		{
			((uint32_t*) m_foundPw)[0] = FROM_BIG_ENDIAN_32(m_h_foundPw[0]);
			((uint32_t*) m_foundPw)[1] = FROM_BIG_ENDIAN_32(m_h_foundPw[1]);
			((uint32_t*) m_foundPw)[2] = FROM_BIG_ENDIAN_32(m_h_foundPw[2]);
			((uint32_t*) m_foundPw)[3] = FROM_BIG_ENDIAN_32(m_h_foundPw[3]);
			m_foundPw[pwLen] = 0;
			return 1;
		}
	}
	return 0;
}

void GpuCracker::cleanUp()
{
	if (m_foundPw != NULL)
	{
		checkCudaErrors(cudaFreeHost(m_h_foundPw));
		checkCudaErrors(cudaFree    (m_d_foundPw));
		checkCudaErrors(cudaFree    (m_d_shared));
		checkCudaErrors(cudaDeviceReset());
		delete [] m_foundPw;
		m_foundPw = NULL;
	}
}
/*
cudaHostAlloc( ... cudaHostAllocPortable);
cudaHostRegister( ... cudaHostRegisterPortable);


cudaStream_t stream[2];
float* hostPtr;

// create
for (int i = 0; i < 2; ++i)
{
	cudaStreamCreate(&stream[i]);
}
cudaMallocHost(&hostPtr, 2 * size);

// use
for (int i = 0; i < 2; ++i)
{
	cudaMemcpyAsync                     (inputDevPtr  + i * size, hostPtr      + i * size, size, cudaMemcpyHostToDevice, stream[i]);
	MyKernel<<<100, 512, 0, stream[i]>>>(outputDevPtr + i * size, inputDevPtr  + i * size, size);
	cudaMemcpyAsync                     (hostPtr      + i * size, outputDevPtr + i * size, size, cudaMemcpyDeviceToHost, stream[i]);
}

// clean up
for (int i = 0; i < 2; ++i)
{
	cudaStreamDestroy(stream[i]);
}
*/
